/*

  Copyright (C) 2007, Yusuke Sakumoto.
  All rights reserved.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include "fLink.h"
#include "fTcpSender.h"
#include <sstream>
#include <cstdlib>
#include <cmath>

void
FLink::add_flow (FTcpSender * flow)
{
  flows.push_back (flow);
  sum_pkt_all_flow[flow] = 0.0;
  pre_time_all_flow[flow] = 0.0;
}

void
FLink::init_link ()
{
  double timeout_time;
  if (link_prop_delay > 0.0)
    timeout_time = link_prop_delay + 1.0;
  else
    timeout_time = 1.0;

  int i = 0;
  for (std::list < FTcpSender * >::iterator it = flows.begin ();
       it != flows.end (); it++)
    {
      flow_states[(*it)].flow_rate.set_timeout (timeout_time);
      flow_states[(*it)].queue_delay.set_timeout (timeout_time);
      flow_states[(*it)].noloss_prob.set_timeout (timeout_time);

      flow_states[(*it)].flow_rate.put_state (link_prop_delay, 0.0);
      flow_states[(*it)].queue_delay.put_state (link_prop_delay, 0.0);
      flow_states[(*it)].noloss_prob.put_state (0.0, 1.0);
      flow_states[(*it)].noloss_prob.put_state (link_prop_delay, 1.0);
      i++;
    }

}

void
FLink::put_state (FTcpSender * flow, double time, double rate, double delay,
		  double noloss, int prev_prev_node_id)
{
  FlowState *state = &flow_states[flow];
  (state->flow_rate).put_state (time, rate);
  (state->queue_delay).put_state (time, delay);
  (state->noloss_prob).put_state (time, noloss);

  if (output->get_print_flag_trace_all () && (prev_node_id >= 0))
    {
      print_trace_all_flow (flow, prev_prev_node_id, time, rate, delay,
			    1.0 - queue->get_loss_prob ());
    }

  if (output->get_print_flag_namtrace_all () && (prev_node_id >= 0))
    {
      print_namtrace_all_flow (flow, prev_prev_node_id, time, rate, delay,
			       1.0 - queue->get_loss_prob ());
    }
}

void
FLink::print_stat ()
{
  if (!output)
    return;

  std::stringstream ss;
  ss << "\nLink configuration"
    << "\nAddress :" << this
    << "\nPrev & Next : " << prev_node << "," << next_node << this->
    get_link_id () << "\nLink delay : " << link_prop_delay <<
    "\nBand Width : " << BW << "\nTraveling flows : " << std::endl;
  output->output_f (ss.str ());
  for (std::list < FTcpSender * >::iterator it = flows.begin ();
       it != flows.end (); it++)
    {
      ss.str ("");
      ss << "\t" << (*it) << (*it)->get_flow_id () << std::endl;
      output->output_f (ss.str ());
    }
  output->output_f (std::string ("\n"));

}

void
FLink::print ()
{
  if (!output)
    return;

  std::stringstream ss;
  ss << "\nLINK STATE"
    << "\nAddress : " << this
    << "\nPrev & Next : " << prev_node_id << " -> " << next_node_id << std::
    endl;
  output->output_f (ss.str ());
  int i = 0;
  for (std::list < FTcpSender * >::iterator it = flows.begin ();
       it != flows.end (); it++)
    {
      ss.str ("");
      ss << "\t flow  : " << (*it)
	<< " (" << (*it)->get_src_node_id () << "," << (*it)->
	get_src_port_id () << ") -> (" << (*it)->
	get_dst_node_id () << "," << (*it)->
	get_dst_port_id () << ")" << std::endl;
      output->output_f (ss.str ());
      i++;
    }
  output->output_f (std::string ("\n"));

}

void
FLink::print_trace_all_flow (FTcpSender * flow, int prev_prev_node_id,
			     double time, double rate, double delay,
			     double noloss)
{
  double sum_pkt = sum_pkt_all_flow[flow];
  double pre_time = pre_time_all_flow[flow];
  double drop;
  double delta;
  int cnt;
  sum_pkt += rate * (time - pre_time);
  delta = (time - pre_time) / std::floor (sum_pkt);
  cnt = 0;

  std::stringstream ss;

  while (sum_pkt > 1.0)
    {
      if (prev_prev_node_id >= 0)
	{
	  ss << "r " << (pre_time + delta * cnt + link_prop_delay)
	    << " " << prev_prev_node_id
	    << " " << prev_node_id
	    << " tcp 1000 ------- " << flow->get_flow_ID ()
	    << " " << flow->get_src_node_id () << "." << flow->
	    get_src_port_id () << " " << flow->
	    get_dst_node_id () << "." << flow->
	    get_dst_port_id () << " " << 0 << " " << 0 << std::endl;
	}
      if (next_node_id >= 0)
	{
	  drop = (double) rand () / 0xffffffff;
	  if (drop > noloss)
	    {
	      ss << "d " << (pre_time + delta * cnt + link_prop_delay)
		<< " " << prev_node_id
		<< " " << next_node_id
		<< " tcp 1000 ------- " << flow->get_flow_ID ()
		<< " " << flow->get_src_node_id () << "." << flow->
		get_src_port_id () << " " << flow->
		get_dst_node_id () << "." << flow->
		get_dst_port_id () << " " << 0 << " " << 0 << std::endl;
	    }
	  else
	    {
	      ss << "+ " << (pre_time + delta * cnt + link_prop_delay)
		<< " " << prev_node_id
		<< " " << next_node_id
		<< " tcp 1000 ------- " << flow->get_flow_ID ()
		<< " " << flow->get_src_node_id () << "." << flow->
		get_src_port_id () << " " << flow->
		get_dst_node_id () << "." << flow->
		get_dst_port_id () << " " << 0 << " " << 0 << std::endl;
	      ss << "- " << pre_time + delta * cnt + delay +
		link_prop_delay << " " << prev_node_id << " " << next_node_id
		<< " tcp 1000 ------- " << flow->
		get_flow_ID () << " " << flow->
		get_src_node_id () << "." << flow->
		get_src_port_id () << " " << flow->
		get_dst_node_id () << "." << flow->
		get_dst_port_id () << " " << 0 << " " << 0 << std::endl;
	    }
	}
      cnt++;
      sum_pkt -= 1.0;
    }
  output->output_f (ss.str ());
  pre_time_all_flow[flow] = time;
  sum_pkt_all_flow[flow] = sum_pkt;

}

void
FLink::print_namtrace_all_flow (FTcpSender * flow, int prev_prev_node_id,
				double time, double rate, double delay,
				double noloss)
{
  double sum_pkt = sum_pkt_all_flow[flow];
  double pre_time = pre_time_all_flow[flow];
  double drop;
  double delta;
  int cnt;
  sum_pkt += rate * (time - pre_time);
  delta = (time - pre_time) / std::floor (sum_pkt);
  cnt = 0;

  std::stringstream ss;

  while (sum_pkt > 1.0)
    {
      if (prev_prev_node_id >= 0)
	{
	  ss << "h -t " << (pre_time + delta * cnt) / 1000.0
	    << " -s " << prev_prev_node_id
	    << " -d " << prev_node_id
	    << " -p tcp -e 1000 -c 0 -i 0 -a 0 -x {" << flow->
	    get_src_node_id () << "." << flow->
	    get_src_port_id () << " " << flow->
	    get_dst_node_id () << "." << flow->
	    get_dst_port_id () << " -1 ------- null}" << std::endl;
	  ss << "r -t " << (pre_time + delta * cnt +
			    link_prop_delay) /
	    1000.0 << " -s " << prev_prev_node_id << " -d " << prev_node_id <<
	    " -p tcp -e 1000 -c 0 -i 0 -a 0 -x {" << flow->
	    get_src_node_id () << "." << flow->
	    get_src_port_id () << " " << flow->
	    get_dst_node_id () << "." << flow->
	    get_dst_port_id () << " -1 ------- null}" << std::endl;
	}
      if (next_node_id >= 0)
	{
	  drop = (double) rand () / 0xffffffff;
	  if (drop > noloss)
	    {
	      ss << "d -t " << (pre_time + delta * cnt +
				link_prop_delay) /
		1000.0 << " -s " << prev_node_id << " -d " << next_node_id <<
		" -p tcp -e 1000 -c 0 -i 0 -a 0 -x {" << flow->
		get_src_node_id () << "." << flow->
		get_src_port_id () << " " << flow->
		get_dst_node_id () << "." << flow->
		get_dst_port_id () << " 0 ------- null}" << std::endl;
	    }
	  else
	    {
	      ss << "+ -t " << (pre_time + delta * cnt +
				link_prop_delay) /
		1000.0 << " -s " << prev_node_id << " -d " << next_node_id <<
		" -p tcp -e 1000 -c 0 -i 0 -a 0 -x {" << flow->
		get_src_node_id () << "." << flow->
		get_src_port_id () << " " << flow->
		get_dst_node_id () << "." << flow->
		get_dst_port_id () << " 0 ------- null}" << std::endl;
	      ss << "- -t " << pre_time + delta * cnt + delay +
		link_prop_delay << " -s " << prev_node_id << " -d " <<
		next_node_id << " -p tcp -e 1000 -c 0 -i 0 -a 0 -x {" <<
		flow->get_src_node_id () << "." << flow->
		get_src_port_id () << " " << flow->
		get_dst_node_id () << "." << flow->
		get_dst_port_id () << " 0 ------- null}" << std::endl;
	    }
	}
      cnt++;
      sum_pkt -= 1.0;
    }
  output->output_f (ss.str ());
  pre_time_all_flow[flow] = time;
  sum_pkt_all_flow[flow] = sum_pkt;
}
