/*

  Copyright (C) 2007, Yusuke Sakumoto.
  All rights reserved.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include "fQueue.h"
#include <sstream>
#include <cmath>

FQueue::FQueue (const int queue_id_, FLink * const link_,
		const double queue_limit_, FSimOutput * const output_,
		const int eq_num_):
queue_id (queue_id_),
link (link_),
queue_limit (queue_limit_),
eq_num (eq_num_),
output (output_),
queue_length (0.0),
arrival_rate (0.0),
departure_rate (0.0),
loss_prob (0.0)
{
  C = link->get_BW ();
}

inline bool
FQueue::is_dummy_queue ()
{
  return (link->get_prev_node_id () * link->get_next_node_id () < 0);
}

FDropTailQueue::FDropTailQueue (const int queue_id_, FLink * const link_, const double queue_limit_, FSimOutput * const output_):FQueue (queue_id_, link_, queue_limit_, output_,
	1)
{
}

FRedQueue::FRedQueue (const int queue_id_, FLink * const link_,
		      const double queue_limit_, const double minth_,
		      const double maxth_, const double maxp_,
		      const double alpha_, FSimOutput * const output_):
FQueue (queue_id_, link_, queue_limit_, output_, 2),
minth (minth_),
maxth (maxth_),
maxp (maxp_),
alpha (alpha_)
{
}

void
FDropTailQueue::put_state (const double time, int *const ins,
			   double *const state, const double arrival_rate_)
{
  double queue_length_ = state[*ins];
  double queue_limit_ = get_queue_limit ();
  set_arrival_rate (arrival_rate_);

  if (queue_length_ > queue_limit_)
    {
      set_loss_prob (1.0);
      set_queue_length (queue_limit_);
    }
  else
    {
      set_loss_prob (0.0);
      set_queue_length (std::max (0.0, queue_length_));
    }

  set_departure_rate (std::
		      min (get_C (),
			   arrival_rate_ * (1.0 - get_loss_prob ())));
  *ins += get_eq_num ();

  print (time);
}

void
FRedQueue::put_state (const double time, int *const ins, double *const state,
		      const double arrival_rate_)
{
  set_arrival_rate (arrival_rate_);
  set_queue_length (std::
		    max (0.0, std::min (get_queue_limit (), state[*ins])));
  avg_queue_length =
    std::max (0.0, std::min (get_queue_limit (), state[*ins + 1]));

  double pb;
  if (avg_queue_length < minth)
    pb = 0.0;
  else
    {
      if (avg_queue_length < maxth)
	pb = maxp / (maxth - minth) * (avg_queue_length - minth);
      else
	{
	  if (avg_queue_length < 2.0 * maxth)
	    pb = (1.0 - maxp) / maxth * avg_queue_length - (1.0 - 2.0 * maxp);
	  else
	    {
	      pb = 1.0;
	    }
	}
    }

  set_loss_prob (2.0 * pb / (1.0 + pb));
  set_departure_rate (std::
		      min (get_C (),
			   arrival_rate_ * (1.0 - get_loss_prob ())));

  *ins += get_eq_num ();
  print (time);
}

std::string FQueue::get_queue_id ()
{
  return link->get_link_id ();
}

std::string FDropTailQueue::get_queue_setting ()
{
  std::stringstream ss;
  ss << "DT" << get_queue_id () << "( " << this << "," << get_C () << "," <<
    get_queue_limit () << ")";
  return ss.str ();
}

void
FDropTailQueue::print_stat ()
{
  FSimOutput *const output_ = get_output ();
  if (!output_)
    return;
  std::stringstream ss;
  ss << "DropTail Queue configuration"
    << "\nAddress : " << this
    << "\nC : " << get_C ()
    << "\nqueue limit : " << get_queue_limit () << std::endl;
  output_->output_f (ss.str ());
}

void
FDropTailQueue::print_avg ()
{
  double avg_queue_;
  double avg_arrival_;
  double avg_departure_;
  double avg_loss_;
  FSimOutput *const output_ = get_output ();
  if (!output_)
    return;

  if (is_dummy_queue ())
    return;
  get_avg (&avg_queue_, &avg_arrival_, &avg_departure_, &avg_loss_);

  std::stringstream ss;
  ss << "DT" << get_queue_id ()
    << " a " << avg_arrival_
    << " d " << avg_departure_
    << " q " << avg_queue_ << " p " << avg_loss_ << std::endl;
  output_->output_f (ss.str ());
}

void
FDropTailQueue::print (double time)
{

  FSimOutput *const output_ = get_output ();
  if (!output_)
    return;
  if (is_dummy_queue ())
    return;

  if (output_->get_print_flag_dyna ())
    {
      std::stringstream ss;
      ss << "DT" << get_queue_id ()
	<< " t " << time / 1000
	<< " a " << get_arrival_rate ()
	<< " d " << get_departure_rate ()
	<< " q " << get_queue_length ()
	<< " p " << get_loss_prob () << std::endl;
      output_->output_f (ss.str ());
    }
  if (output_->get_print_flag_avg ())
    {
      update_avg (time);
    }

}

std::string FRedQueue::get_queue_setting ()
{
  std::stringstream ss;
  ss << "RED" << get_queue_id ()
    << " (" << this
    << "," << get_C ()
    << "," << get_queue_limit ()
    << "," << minth << "," << maxth << "," << maxp << "," << alpha << ")";
  return ss.str ();
}

void
FRedQueue::print_stat ()
{
  FSimOutput *const output_ = get_output ();

  if (!output_)
    return;
  std::stringstream ss;
  ss << "Red Queue configuration"
    << "\nAddress : " << this
    << "\nC : " << get_C ()
    << "\nmin threshold : " << minth
    << "\nmax threshold : " << maxth
    << "\nmax prob. : " << maxp
    << "\nalpha : " << alpha
    << "\nqueue limit : " << get_queue_limit () << std::endl;
  output_->output_f (ss.str ());
}

void
FRedQueue::print_avg ()
{
  double avg_queue_;
  double avg_arrival_;
  double avg_departure_;
  double avg_loss_;
  double avg_avg_queue_;
  FSimOutput *const output_ = get_output ();
  if (!output_)
    return;
  if (is_dummy_queue ())
    return;

  FRedQueue::get_avg (&avg_queue_, &avg_arrival_, &avg_departure_, &avg_loss_,
		      &avg_avg_queue_);

  std::stringstream ss;
  ss << "RED" << get_queue_id ()
    << " a " << avg_arrival_
    << " d " << avg_departure_
    << " q " << avg_queue_
    << " avgq " << avg_avg_queue_ << " p " << avg_loss_ << std::endl;
  output_->output_f (ss.str ());
}

void
FRedQueue::print (double time)
{
  FSimOutput *const output_ = get_output ();
  if (!output_)
    return;

  if (output_->get_print_flag_dyna ())
    {
      std::stringstream ss;
      ss << "RED" << get_queue_id ()
	<< " t " << time / 1000.0
	<< " a " << get_arrival_rate ()
	<< " d " << get_departure_rate ()
	<< " q " << get_queue_length ()
	<< " avgq " << avg_queue_length
	<< " p " << get_loss_prob () << std::endl;
      output_->output_f (ss.str ());
    }
  if (output_->get_print_flag_avg ())
    {
      update_avg (time);
    }
}
