/*

  Copyright (C) 2007, Yusuke Sakumoto.
  All rights reserved.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#ifndef __FQueue
#define __FQueue

#include "common.h"
#include "fNode.h"
#include "fLink.h"
#include "fRouter.h"
#include "fTcpSender.h"
#include "fSimOutput.h"

class FQueue:public FNode
{
  // identification parameter
  const int queue_id;
  class FLink *const link;

  // setting parameter
  const double queue_limit;
  double C;
  const int eq_num;
  FSimOutput *const output;
  AvgData avg_queue, avg_arrival, avg_departure, avg_loss;

  // modeling parameter
  double queue_length;

  // other parameter
  double arrival_rate;
  double departure_rate;
  double loss_prob;

public:
    FQueue (const int queue_id_, FLink * const link_,
	    const double queue_limit_, FSimOutput * const output_,
	    const int eq_num_);

  bool is_dummy_queue ();
  double queue_derivs (double queue_length_, double arrival_rate_)
  {
    if (queue_length_ > 0.0)
      return arrival_rate_ - get_C ();
    else
      return std::max (0.0, (arrival_rate_ - get_C ()));
  }
  void update_avg (const double time)
  {
    avg_queue.update_avg (time, queue_length);
    avg_arrival.update_avg (time, arrival_rate);
    avg_departure.update_avg (time, departure_rate);
    avg_loss.update_avg (time, loss_prob);
  }
  void get_avg (double *avg_queue_, double *avg_arrival_,
		double *avg_departure_, double *avg_loss_)
  {
    *avg_queue_ = avg_queue.get_avg ();
    *avg_arrival_ = avg_arrival.get_avg ();
    *avg_departure_ = avg_departure.get_avg ();
    *avg_loss_ = avg_loss.get_avg ();
  }

  // set function
  void set_arrival_rate (const double arrival_rate_)
  {
    arrival_rate = arrival_rate_;
  }
  void set_departure_rate (const double departure_rate_)
  {
    departure_rate = departure_rate_;
  }
  void set_loss_prob (const double loss_prob_)
  {
    loss_prob = loss_prob_;
  }
  void set_queue_length (const double queue_length_)
  {
    queue_length = queue_length_;
  }
  virtual void set_max_min (int *ins, double *max_values,
			    double *min_values) = 0;

  // get function
  std::string get_queue_id ();
  double get_queue_limit ()
  {
    return queue_limit;
  }
  double get_C ()
  {
    return C;
  }
  int get_eq_num ()
  {
    return eq_num;
  }
  FSimOutput *const get_output ()
  {
    return output;
  }
  double get_queue_length ()
  {
    return queue_length;
  }
  double get_arrival_rate ()
  {
    return arrival_rate;
  }
  double get_loss_prob ()
  {
    return loss_prob;
  }
  double get_departure_rate ()
  {
    return departure_rate;
  }
  double get_queuing_delay ()
  {
    return queue_length / C;
  }
  virtual void get_derivs (double *const derivs, int *const ins,
			   double *const state, const double arrival_rate_) =
    0;
  virtual void get_state (double *const state, int *const ins) = 0;
  virtual std::string get_queue_setting () = 0;

  // print function
  virtual void print_stat () = 0;
  virtual void print (double time) = 0;
  virtual void print_avg () = 0;

  virtual void put_state (const double time, int *const ins,
			  double *const state, const double arrival_rate_) =
    0;

};

class FDummyQueue:public FQueue
{
public:
  FDummyQueue (const int queue_id_, FLink * const link_):FQueue (queue_id_,
								 link_, 0,
								 NULL, 0)
  {
  }
  void get_derivs (double *const derivs, int *const ins, double *const state,
		   const double arrival_rate_)
  {
  }
  void get_state (double *const state, int *const ins)
  {
  };
  std::string get_queue_setting ()
  {
    return "";
  }

  // print function
  void print_stat ()
  {
  }
  void print (double time)
  {
  }
  void print_avg ()
  {
  }
  void set_max_min (int *ins, double *max_values, double *min_values)
  {
  }
  void put_state (const double time, int *const ins, double *const state,
		  const double arrival_rate_)
  {
    set_arrival_rate (arrival_rate_);
    set_departure_rate (arrival_rate_);
  }
};

class FDropTailQueue:public FQueue
{
public:
  FDropTailQueue (const int queue_id_,
		  FLink * const link_, const double queue_limit_,
		  FSimOutput * const output_);

    std::string get_queue_setting ();
  void get_derivs (double *const derivs, int *const ins, double *const state,
		   const double arrival_rate_)
  {
    derivs[*ins] = queue_derivs (state[*ins], arrival_rate_);
    *ins += get_eq_num ();
  }
  void get_state (double *const state, int *const ins)
  {
    state[*ins] = get_queue_length ();
    *ins += get_eq_num ();
  }
  void put_state (const double time, int *const ins, double *const state,
		  const double arrival_rate_);
  void print_stat ();
  void print (double time);
  void set_max_min (int *ins, double *max_values, double *min_values)
  {
    max_values[*ins] = get_queue_limit ();
    min_values[*ins] = 0.0;
    *ins += get_eq_num ();
  }

  void print_avg ();
};

class FRedQueue:public FQueue
{
  // control parameter
  const double minth, maxth, maxp, alpha;

  // modeling parameter
  double avg_queue_length;
  AvgData avg_avg_queue;
public:

    FRedQueue (const int queue_id_, FLink * const link_,
	       const double queue_limit_, const double minth_,
	       const double maxth_, const double maxp_, const double alpha_,
	       FSimOutput * const output_);

    std::string get_queue_setting ();
  void get_derivs (double *const derivs, int *const ins, double *const state,
		   const double arrival_rate_)
  {
    derivs[*ins] = queue_derivs (state[*ins], arrival_rate_);
    derivs[*ins + 1] = avg_queue_derivs (state[*ins], state[*ins + 1]);
    *ins += get_eq_num ();
  }
  void get_state (double *const state, int *const ins)
  {
    state[*ins] = get_queue_length ();
    state[*ins + 1] = avg_queue_length;
    *ins += get_eq_num ();
  }
  void put_state (const double time, int *const ins, double *const state,
		  const double arrival_rate_);
  void print_stat ();
  void print (double time);

  double avg_queue_derivs (double queue_length_, double avg_queue_)
  {
    return alpha * get_C () * (queue_length_ - avg_queue_);
  }

  void update_avg (const double time)
  {
    FQueue::update_avg (time);
    avg_avg_queue.update_avg (time, avg_queue_length);
  }
  void get_avg (double *avg_queue_, double *avg_arrival_,
		double *avg_departure_, double *avg_loss_,
		double *avg_avg_queue_)
  {
    FQueue::get_avg (avg_queue_, avg_arrival_, avg_departure_, avg_loss_);
    *avg_avg_queue_ = avg_avg_queue.get_avg ();
  }
  void print_avg ();

  void set_max_min (int *ins, double *max_values, double *min_values)
  {
    max_values[*ins] = get_queue_limit ();
    max_values[*ins + 1] = get_queue_limit ();
    min_values[*ins] = 0.0;
    min_values[*ins + 1] = 0.0;
    *ins += get_eq_num ();
  }

};
#endif
