/*

  Copyright (C) 2007, Yusuke Sakumoto.
  All rights reserved.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include "fRouter.h"
#include "fQueue.h"
#include "fLink.h"
#include <sstream>
#include <cmath>

void
FRouter::put_state (const double time, int *const ins, double *const state)
{
  double arrival_rate_, departure_rate_;
  output_links_arrival_rate.init_zero ();

  for (std::list < Route >::iterator it = route_table.begin ();
       it != route_table.end (); it++)
    output_links_arrival_rate.add_rate ((*it).output_link,
					((*it).input_link)->get_rate (time,
								      (*it).
								      flow));

  FQueue *queue;
  double rate, delay, noloss;
  for (std::list < FLink * >::iterator it = out_links.begin ();
       it != out_links.end (); it++)
    {
      arrival_rate_ = output_links_arrival_rate.get_rate ((*it));
      queue = (*it)->get_queue ();
      queue->put_state (time, ins, state, arrival_rate_);
      departure_rate_ = queue->get_departure_rate ();
      for (std::list < Route >::iterator route = route_table.begin ();
	   route != route_table.end (); route++)
	{
	  if ((*route).output_link == (*it))
	    {
	      if (arrival_rate_ > 0.0)
		{
		  ((*route).input_link)->get_state (time, (*route).flow,
						    &rate, &delay, &noloss);
		  ((*route).output_link)->put_state ((*route).flow, time,
						     rate / arrival_rate_ *
						     departure_rate_,
						     delay +
						     queue->
						     get_queuing_delay (),
						     noloss * (1.0 -
							       queue->
							       get_loss_prob
							       ()),
						     (*route).input_link->
						     get_prev_node_id ());
		}
	      else
		{
		  ((*route).input_link)->get_state (time, (*route).flow,
						    &rate, &delay, &noloss);
		  ((*route).output_link)->put_state ((*route).flow, time, 0.0,
						     delay +
						     queue->
						     get_queuing_delay (),
						     noloss * (1.0 -
							       queue->
							       get_loss_prob
							       ()),
						     (*route).input_link->
						     get_prev_node_id ());
		}
	    }
	}
    }
}


int
FRouter::get_eq_num ()
{
  int eq_num_ = 0;
  FQueue *queue;
  for (std::list < FLink * >::iterator it = out_links.begin ();
       it != out_links.end (); it++)
    {
      queue = (*it)->get_queue ();
      eq_num_ += queue->get_eq_num ();
    }
  return eq_num_;
}

void
FRouter::get_all_queue_derivs (const double time, double *const derivs,
			       int *const ins, double *const state)
{
  double arrival_rate_, departure_rate_;

  output_links_arrival_rate.init_zero ();
  for (std::list < Route >::iterator it = route_table.begin ();
       it != route_table.end (); it++)
    output_links_arrival_rate.add_rate ((*it).output_link,
					((*it).input_link)->get_rate (time,
								      (*it).
								      flow));

  FQueue *queue;
  for (std::list < FLink * >::iterator it = out_links.begin ();
       it != out_links.end (); it++)
    {
      arrival_rate_ = output_links_arrival_rate.get_rate ((*it));
      queue = (*it)->get_queue ();
      queue->get_derivs (derivs, ins, state, arrival_rate_);
    }
}
void
FRouter::add_route (FTcpSender * const flow, FLink * const prev_link,
		    FLink * const next_link)
{
  Route new_route;

  new_route.flow = flow;
  new_route.input_link = prev_link;
  new_route.output_link = next_link;

  route_table.push_back (new_route);

  for (std::list < FLink * >::iterator it = out_links.begin ();
       it != out_links.end (); it++)
    {
      if ((*it) == next_link)
	return;
    }

  output_links_arrival_rate.add_link (next_link);
  out_links.push_back (next_link);

}

void
FRouter::print_avg ()
{
  FQueue *queue;
  for (std::list < FLink * >::iterator it = out_links.begin ();
       it != out_links.end (); it++)
    {
      queue = (*it)->get_queue ();
      queue->print_avg ();
    }
}
void
FRouter::print_stat ()
{
  std::stringstream ss;
  ss << "Router configuration"
    << "\nAddress : " << this
    << "\nRouter ID : " << router_id
    << "\n# of flows : " << route_table.size () << "\nRouting table : \n";
  output->output_f (ss.str ());

  FQueue *queue;
  for (std::list < Route >::iterator it = route_table.begin ();
       it != route_table.end (); it++)
    {
      queue = ((*it).output_link)->get_queue ();
      ss.str ("");
      ss << "\tRoute " << (*it).flow << ((*it).flow)->get_flow_id ()
	<< " from " << (*it).input_link << ((*it).input_link)->get_link_id ()
	<< " to " << (*it).output_link << ((*it).output_link)->get_link_id ()
	<< " Q " << queue->get_queue_setting () << std::endl;
      output->output_f (ss.str ());
    }
  output->output_f (std::string ("\n"));
}

void
FRouter::set_max_min (int *ins, double *max_values, double *min_values)
{
  FQueue *queue;
  for (std::list < FLink * >::iterator it = out_links.begin ();
       it != out_links.end (); it++)
    {
      queue = (*it)->get_queue ();
      queue->set_max_min (ins, max_values, min_values);
    }
}
