/*

  Copyright (C) 2007, Yusuke Sakumoto.
  All rights reserved.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include "fSimEngine.h"
#include <fstream>
#include <string>
#include <sstream>
#include <iomanip>
#include "fSimInput.h"
#include <cmath>

std::vector < std::string > tokenize (const std::string & input,
				      const char &dim);

std::vector < std::string > tokenize (const std::string & input,
				      const char &dim)
{
  std::vector < std::string > tokens;
  int
    head = 0;
  int
    i = 0;

  for (std::string::const_iterator it = input.begin (); it != input.end ();
       it++)
    {
      if ((*it) == dim)
	{
	  tokens.push_back (input.substr (head, i - head));
	  head = i + 1;
	}
      i++;
    }

  tokens.push_back (input.substr (head, input.size () - head));

  return tokens;
}

FSimEngine::~FSimEngine ()
{
  if (ytemp)
    delete[]ytemp;
  if (ysav)
    delete[]ysav;
  if (dysav)
    delete[]dysav;
  if (dysav_mid)
    delete[]dysav_mid;

  if (derivs1)
    delete[]derivs1;
  if (derivs2)
    delete[]derivs2;
  if (derivs3)
    delete[]derivs3;
  if (state1)
    delete[]state1;
  if (state2)
    delete[]state2;
  if (state3)
    delete[]state3;

}

void
FSimEngine::print_nam_header (std::ofstream & out)
{
  out << "V -t * -v 1.0a5 -a 0\n"
    << "A -t * -n 1 -p 0 -o 0xffffffff -c 31 -a 1\n"
    << "A -t * -h 1 -m 2147483647 -s 0\n";
  router_list.print_nam_header (out);
  link_list.print_nam_header (out);
}

void
FSimEngine::set_output (char *output_file_name, char *output_format_name)
{
  bool print_flag_dyna = false;
  bool print_flag_avg = false;
  bool print_flag_trace_all = false;
  bool print_flag_namtrace_all = false;

  if (!strcmp (output_format_name, "fsim"))
    {
      print_flag_dyna = true;
    }
  else if (!strcmp (output_format_name, "fsim_avg"))
    {
      print_flag_avg = true;
    }
  else if (!strcmp (output_format_name, "trace_all"))
    {
      print_flag_trace_all = true;
    }
  else if (!strcmp (output_format_name, "namtrace_all"))
    {
      print_flag_namtrace_all = true;
    }
  else
    {
      std::cerr << "Wrong format name : " << output_format_name << std::endl;
      std::exit (1);
    }
  output = new FSimOutput (output_file_name, print_flag_dyna, print_flag_avg,
			   print_flag_trace_all, print_flag_namtrace_all);
}

void
FSimEngine::read_input (char *setting_file_name, char *topology_file_name)
{
  std::ifstream in_set (setting_file_name);

  if (!in_set)
    {
      std::cerr << "Cannot open file : " << setting_file_name << std::endl;
      std::exit (1);
    }
  double minth, maxth, maxp, alpha, queue_limit, max_window_size;
  int max_node_num;

  // read parameter setting
  while (!in_set.eof ())
    {
      std::string in_buffer;
      getline (in_set, in_buffer);
      std::vector < std::string > token = tokenize (in_buffer, ' ');
      if (token[0] == "minth")
	{
	  minth = atof (token[1].c_str ());
	}
      else if (token[0] == "maxth")
	{
	  maxth = atof (token[1].c_str ());
	}
      else if (token[0] == "alpha")
	{
	  alpha = atof (token[1].c_str ());
	}
      else if (token[0] == "init_delta")
	{
	  init_delta = atof (token[1].c_str ());
	}
      else if (token[0] == "simulation_time")
	{
	  sim_time = atof (token[1].c_str ());
	}
      else if (token[0] == "queue_limit")
	{
	  queue_limit = atof (token[1].c_str ());
	}
      else if (token[0] == "maxp")
	{
	  maxp = atof (token[1].c_str ());
	}
      else if (token[0] == "max_node_num")
	{
	  max_node_num = atoi (token[1].c_str ());
	}
      else if (token[0] == "min_step")
	{
	  MINSTEP = atof (token[1].c_str ());
	}
      else if (token[0] == "max_step")
	{
	  MAXSTEP = atof (token[1].c_str ());
	}
      else if (token[0] == "tiny")
	{
	  TINY = atof (token[1].c_str ());
	}
      else if (token[0] == "pgrow")
	{
	  PGROW = atof (token[1].c_str ());
	}
      else if (token[0] == "pshrnk")
	{
	  PSHRNK = atof (token[1].c_str ());
	}
      else if (token[0] == "fcor")
	{
	  FCOR = atof (token[1].c_str ());
	}
      else if (token[0] == "safety")
	{
	  SAFETY = atof (token[1].c_str ());
	}
      else if (token[0] == "errcon")
	{
	  ERRCON = atof (token[1].c_str ());
	}
      else if (token[0] == "eps")
	{
	  EPS = atof (token[1].c_str ());
	}
      else if (token[0] == "init_throughput")
	{
	  init_thr = atof (token[1].c_str ());
	}
      else if (token[0] == "max_window_size")
	{
	  max_window_size = atof (token[1].c_str ());
	}
    }

#if __DEBUG
  std::stringstream ss;
  ss << "System parameter setting\n"
    << "\nSimulation time : " << sim_time
    << "\nMinimum threshold : " << minth
    << "\nMaximum threshold : " << maxth
    << "\nMax provability : " << maxp
    << "\nAlpha : " << alpha
    << "\nQueue limit : " << queue_limit
    << "\nInitial Delta : " << init_delta
    << "\nInitial throughput : " << init_thr << std::endl << std::endl;
  output->output_f (ss.str ());
#endif

  std::ifstream in (topology_file_name);

  if (!in)
    {
      std::cerr << "Cannnot open file : " << topology_file_name << std::endl;
      std::exit (1);
    }
  int *port_manager = new int[max_node_num];
  for (int i = 0; i < max_node_num; i++)
    port_manager[i] = 0;

  int queue_id = 0;		// more than 0 and equal to 0
  int flow_id = -1;		// less than 0

  while (!in.eof ())
    {
      std::string str;
      std::getline (in, str, '\n');	// read line

      std::vector < std::string > tokens = tokenize (str, '\t');
      if (tokens.size () < 2)
	break;			// fomat error
      std::vector < std::string > flow_tok = tokenize (tokens[0], '-');
      std::vector < std::string > route_tok = tokenize (tokens[1], ',');

      // setting new flow
      int src_node_id = atoi (flow_tok[0].c_str ());
      int dst_node_id = atoi (flow_tok[1].c_str ());
      int prev_node_id = flow_id;
      int now_node_id = src_node_id;
      int next_node_id;

      if (src_node_id > max_node_num || dst_node_id > max_node_num)
	{
	  std::cerr << "Node ID is over 'max_node_num'\n";
	  std::cerr << "Please change the value of 'max_node_num'\n";
	  std::exit (1);
	}

      FTcpSender *new_flow = flow_list.add_list (src_node_id, dst_node_id);
      if (new_flow == NULL)
	continue;
      FLink *send_link =
	link_list.add_list (prev_node_id, now_node_id, new_flow);
      send_link->set_output (output);
      send_link->set_link_prop_delay (0.0);
      send_link->set_prev_node (new_flow);
      send_link->set_BW (INF);

      int src_port_id = port_manager[src_node_id - 1]++;
      int dst_port_id = port_manager[dst_node_id - 1]++;

      new_flow->set_output (output);
      new_flow->set_send_link (send_link);
      new_flow->set_max_window_size (max_window_size);
      new_flow->set_src_port_id (src_port_id);
      new_flow->set_dst_port_id (dst_port_id);
      new_flow->set_init_thr (init_thr);
      new_flow->set_flow_ID (-flow_id);

      FLink *prev_link = send_link;
      FLink *next_link;

      FRouter *new_router;

      std::string next_queue_type;

      FQueue *new_queue;

      double link_prop_delay;
      double sum_link_prop_delay = 0.0;
      // double sum_link_delay = 0.0;
      double BW;
      int cnt_route;

      std::vector < std::string > attr_next = tokenize (route_tok[0], ':');
      std::vector < std::string > attr;
      for (cnt_route = 0; cnt_route < route_tok.size () - 1; cnt_route++)
	{
	  attr = attr_next;

	  // get attributions of "now node"
	  BW = atoi (attr[1].c_str ());
	  link_prop_delay = atoi (attr[2].c_str ());
	  next_queue_type = attr[3];

	  sum_link_prop_delay += link_prop_delay;
	  // sum_link_delay += link_delay;
	  attr_next = tokenize (route_tok[cnt_route + 1], ':');
	  next_node_id = atoi (attr_next[0].c_str ());
	  if (cnt_route == 0)
	    new_flow->set_next_node_id (next_node_id);

	  // add next_link(from now_node_id to next_node_id) to new_flow
	  next_link =
	    link_list.add_list (now_node_id, next_node_id, new_flow);
	  next_link->set_link_prop_delay (link_prop_delay);
	  next_link->set_output (output);
	  next_link->set_BW (BW);

	  if (next_link->get_queue () == NULL)
	    {
	      if (next_queue_type == "DropTail")
		{
		  new_queue =
		    new FDropTailQueue (queue_id++, next_link, queue_limit,
					output);
		}
	      else if (next_queue_type == "RED")
		{
		  new_queue =
		    new FRedQueue (queue_id++, next_link, queue_limit, minth,
				   maxth, maxp, alpha, output);
		}
	      next_link->set_queue (new_queue);
	    }
	  else
	    {
	      new_queue = next_link->get_queue ();
	    }

	  // make router(now_node_id)
	  new_router =
	    router_list.add_list (now_node_id, prev_link, next_link,
				  new_flow);
	  new_router->set_output (output);

	  new_flow->add_queue (new_queue);

	  prev_link->set_next_node (new_router);
	  next_link->set_prev_node (new_router);

	  prev_link = next_link;
	  prev_node_id = now_node_id;
	  now_node_id = next_node_id;
	}

      // make sink link
      FLink *sink_link =
	link_list.add_list (dst_node_id, flow_id--, new_flow);
      sink_link->set_link_prop_delay (sum_link_prop_delay);
      sink_link->set_output (output);
      sink_link->set_BW (INF);	// Don't drop an ack packet.
      sink_link->set_queue (new FDummyQueue (queue_id++, sink_link));
      new_router =
	router_list.add_list (dst_node_id, prev_link, sink_link, new_flow);
      new_router->set_output (output);

      new_flow->add_queue (sink_link->get_queue ());

      prev_link->set_next_node (new_router);
      sink_link->set_prev_node (new_router);
      sink_link->set_next_node (new_flow);

      new_flow->set_sink_link (sink_link);
      new_flow->set_prop_delay (2.0 * sum_link_prop_delay);
      new_flow->set_init_rtt (2.0 * sum_link_prop_delay);
    }

#if __DEBUG
  flow_list.print ();
  link_list.print ();
  router_list.print ();
#endif

  flow_num = flow_list.size ();
  router_num = router_list.size ();
}


void
FSimEngine::init (double *state)
{
  int i = 0;
  for (; i < flow_num; i++)
    state[i] = init_thr;

  for (; i < eq_num; i++)
    state[i] = 0.0;

  link_list.init_links ();
  set_max_min ();

  ytemp = new double[eq_num];
  ysav = new double[eq_num];
  dysav = new double[eq_num];
  dysav_mid = new double[eq_num];

  derivs1 = new double[eq_num];
  derivs2 = new double[eq_num];
  derivs3 = new double[eq_num];
  state1 = new double[eq_num];
  state2 = new double[eq_num];
  state3 = new double[eq_num];
}

void
FSimEngine::simulation ()
{
  double actual_delta, next_delta;
  double next_time;
  double now = 0.0;
  double pre_time = now;
  double delta = init_delta;

  double *scal;
  double *derivs;
  double *state;

  // eq_num = flow_num + 4*router_num;
  flow_list.set_eq_offset (0);
  router_list.set_eq_offset (flow_num);

  eq_num = flow_num + router_list.get_eq_num ();

  scal = new double[eq_num];
  derivs = new double[eq_num];
  state = new double[eq_num];

  init (state);
  while (pre_time <= sim_time)
    {
      std::cerr << std::setiosflags (std::ios::fixed) << std::
	setw (8) << std::setprecision (3) << now / 1000.0 << "/" << sim_time /
	1000.0 << "\r";
      get_derivs_all_node (now, state, derivs);

      // scalling
      for (int i = 0; i < eq_num; i++)
	{
	  scal[i] =
	    std::fabs (state[i]) + std::fabs (derivs[i] * init_delta) + TINY;
	  // scal[i] = std::fabs(derivs[i]*init_delta) + TINY;
	}

      // calculate network state
      rkqs (state, derivs, now, delta, &actual_delta, &next_time, &next_delta,
	    scal);

      // save network state 
      flow_list.flow_puts (next_time, state);
      router_list.router_puts (next_time, state);

      step_time.update_avg (next_time, actual_delta);
      pre_time = now;

      delta = next_delta;
      now = next_time;
    }

  if (output->get_print_flag_avg ())
    {
      flow_list.print_avg ();
      router_list.print_avg ();
      std::stringstream ss;
      ss << "avg step time: " << step_time.get_avg () << std::endl;
      output->output_f (ss.str ());
    }
  // clear output buffer
  output->flush ();

  delete[]scal;
  delete[]derivs;
  delete[]state;
}

void
FSimEngine::rk4_all_node (double *state, double *derivs, double now,
			  double delta, double *state_out)
{
  int i;
  double delta_h = delta / 2.0;

  for (i = 0; i < eq_num; i++)
    state1[i] =
      std::min (max_values[i],
		std::max (min_values[i], state[i] + delta * derivs[i] / 2.0));
  get_derivs_all_node (now + delta / 2.0, state1, derivs1);	// state1 -> derivs1

  for (i = 0; i < eq_num; i++)
    state2[i] =
      std::min (max_values[i],
		std::max (min_values[i],
			  state[i] + delta * derivs1[i] / 2.0));
  get_derivs_all_node (now + delta / 2.0, state2, derivs2);	// state2 -> derivs2

  for (i = 0; i < eq_num; i++)
    state3[i] =
      std::min (max_values[i],
		std::max (min_values[i], state[i] + delta * derivs2[i]));
  get_derivs_all_node (now + delta, state3, derivs3);	// state3 -> derivs3

  for (i = 0; i < eq_num; i++)
    state_out[i] =
      std::min (max_values[i],
		std::max (min_values[i],
			  state[i] + delta * (derivs[i] +
					      2.0 * (derivs1[i] +
						     derivs2[i]) +
					      derivs3[i]) / 6.0));

}

void
FSimEngine::rkqs (double *state, double *derivs, double now,
		  double init_delta, double *actual_delta, double *next_time,
		  double *next_delta, double *scal)
{
  double xsav = now;
  double x;
  double h = init_delta;
  double hh;
  double errmax, temp;

#if __ADAPT
  for (int i = 0; i < eq_num; i++)
    {
      ysav[i] = state[i];
      dysav[i] = derivs[i];
    }

  for (;;)
    {
      hh = 0.5 * h;
      // rk4(xsav, xsav+hh) -> ytemp
      rk4_all_node (ysav, dysav, xsav, hh, ytemp);
      x = xsav + hh;
      get_derivs_all_node (x, ytemp, dysav_mid);
      // rk4(xsav+hh, xsav+h) -> state
      rk4_all_node (ytemp, dysav_mid, x, hh, state);

      // rk4(xsav, xsav+h) -> ytemp
      rk4_all_node (ysav, dysav, xsav, h, ytemp);

      errmax = 0.0;
      for (int i = 0; i < eq_num; i++)
	{
	  ytemp[i] = state[i] - ytemp[i];
	  temp = std::fabs (ytemp[i] / scal[i]);
	  if (errmax < temp)
	    errmax = temp;
	}

      errmax /= EPS;

      if (h <= MINSTEP && errmax > 1.0)
	{
	  *actual_delta = h;
	  *next_delta = MINSTEP;
	  break;
	}

      if (errmax <= 1.0)
	{
	  *actual_delta = h;

	  if (errmax > ERRCON)
	    {
	      *next_delta = SAFETY * h * std::exp (PGROW * std::log (errmax));
	    }
	  else
	    {
	      *next_delta = 4.0 * h;
	    }
	  break;
	}

      h = SAFETY * h * std::exp (PSHRNK * std::log (errmax));

    }

  *next_time = now + *actual_delta;
  for (int i = 0; i < eq_num; i++)
    state[i] += ytemp[i] * FCOR;
#else
  rk4_all_node (state, derivs, now, init_delta, ytemp);

  *actual_delta = init_delta;
  *next_delta = init_delta;
  *next_time = now + init_delta;

  for (int i = 0; i < eq_num; i++)
    state[i] = ytemp[i];

#endif
}

void
FSimEngine::get_derivs_all_node (double time, double *state, double *derivs)
{
  flow_list.get_derivs (time, state, derivs);
  router_list.get_derivs (time, state, derivs);
}

void
FSimEngine::set_max_min ()
{
  if (max_values)
    delete[]max_values;
  if (min_values)
    delete[]min_values;

  max_values = new double[eq_num];
  min_values = new double[eq_num];

  flow_list.set_max_min (max_values, min_values);
  router_list.set_max_min (max_values, min_values);
}
