/*

  Copyright (C) 2007, Yusuke Sakumoto.
  All rights reserved.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include "fTcpSender.h"
#include <sstream>
#include <cmath>

FTcpSender::FTcpSender ()
{
  port_id = 0;
  src_node_id = dst_node_id = src_port_id = dst_port_id = 0;
  flow_num = 1;
  rtt.put_state (0.0, init_rtt);
  thr.put_state (0.0, init_thr);
  loss = 0.0;
  init_thr = 0.0;
  max_throughput = 0.0;
  send_link = sink_link = NULL;
}

void
FTcpSender::puts_state (double time, double new_thr)
{

  double rtt_now = sink_link->get_queue_delay (time, this) + init_rtt;
  double loss_now = 1.0 - sink_link->get_noloss_prob (time, this);
  new_thr =
    std::min (max_window_size / rtt_now, std::max (init_thr, new_thr));

  thr.put_state (time, new_thr);
  send_link->put_state (this, time, flow_num * new_thr);

  if (output)
    {
      // print dynamics
      if (output->get_print_flag_dyna ())
	{
	  std::stringstream ss;
	  ss << "TCP" << this->get_flow_id ()
	    << " n " << flow_num
	    << " t " << time / 1000
	    << " r " << new_thr
	    << " loss " << loss_now
	    << " rtt " << rtt_now << " w " << new_thr * rtt_now << std::endl;
	  output->output_f (ss.str ());
	}
      // for printing average
      if (output->get_print_flag_avg ())
	{
	  avg_thr.update_avg (time, new_thr);
	  avg_rtt.update_avg (time, rtt_now);
	  avg_loss.update_avg (time, loss_now);
	}
      // print trace-all
      if (output->get_print_flag_trace_all ())
	{
	  static double sum_pkt = 0.0;
	  static double pre_time = 0.0;
	  double delta;
	  int cnt;
	  sum_pkt += new_thr * flow_num * (time - pre_time);
	  delta = (time - pre_time) / std::floor (sum_pkt);
	  cnt = 0;
	  std::stringstream ss;

	  while (sum_pkt > 1.0)
	    {
	      // departure packet
	      ss << "+ " << (pre_time + delta * cnt) / 1000.0
		<< " " << src_node_id
		<< " " << next_node_id
		<< " tcp 1000 ------- " << flow_ID
		<< " " << src_node_id << "." << src_port_id
		<< " " << dst_node_id << "." << dst_port_id
		<< " " << 0 << " " << 0 << std::endl;
	      ss << "- " << (pre_time + delta * cnt) / 1000.0
		<< " " << src_node_id
		<< " " << next_node_id
		<< " tcp 1000 ------- " << flow_ID
		<< " " << src_node_id << "." << src_port_id
		<< " " << dst_node_id << "." << dst_port_id
		<< " " << 0 << " " << 0 << std::endl;
	      cnt++;
	      sum_pkt -= 1.0;
	    }
	  pre_time = time;
	  output->output_f (ss.str ());
	}
      // print trace-all
      if (output->get_print_flag_namtrace_all ())
	{
	  static double sum_pkt = 0.0;
	  static double pre_time = 0.0;
	  double delta;
	  int cnt;
	  sum_pkt += new_thr * flow_num * (time - pre_time);
	  delta = (time - pre_time) / std::floor (sum_pkt);
	  cnt = 0;
	  std::stringstream ss;

	  while (sum_pkt > 1.0)
	    {
	      // departure packet
	      ss << "+ -t " << (pre_time + delta * cnt) / 1000.0
		<< " -s " << src_node_id
		<< " -d " << next_node_id
		<< " -p tcp -e 1000 -c 0 -i 0 -a 0 -x {" << src_node_id << "."
		<< src_port_id << " " << dst_node_id << "." << dst_port_id <<
		" 0 ------- null}" << std::endl;
	      ss << "- -t " << (pre_time +
				delta * cnt) /
		1000.0 << " -s " << src_node_id << " -d " << next_node_id <<
		" -p tcp -e 1000 -c 0 -i 0 -a 0 -x {" << src_node_id << "." <<
		src_port_id << " " << dst_node_id << "." << dst_port_id <<
		" 0 ------- null}" << std::endl;
	      cnt++;
	      sum_pkt -= 1.0;
	    }
	  pre_time = time;
	  output->output_f (ss.str ());
	}
    }
}

double
FTcpSender::calc_loss (double time)
{
  double p = 1.0;
  for (std::list < FQueue * >::iterator it = queue_list.begin ();
       it != queue_list.end (); it++)
    {
      p *= (1.0 - (*it)->get_loss_prob ());
    }
  loss = 1.0 - p;
  return loss;
}

double
FTcpSender::calc_rtt (double time)
{
  double rtt_ = init_rtt;
  for (std::list < FQueue * >::iterator it = queue_list.begin ();
       it != queue_list.end (); it++)
    {
      rtt_ += (*it)->get_queuing_delay ();
    }

  rtt.put_state (time, rtt_);
  return rtt_;
}

void
FTcpSender::print_stat ()
{

  std::stringstream ss;
  ss << "Tcp configuration"
    << "\nAddress : " << this
    << this->get_flow_id ()
    << "\n# of flows : " << flow_num
    << "\nRound-trip Propagetion delay : " << prop_delay
    << "\nSend link : " << send_link
    << "\nSink link : " << sink_link << "\nRoute : ";

  // output->output_f(ss.str());
  for (std::list < FQueue * >::iterator it = queue_list.begin ();
       it != queue_list.end (); it++)
    {
      ss << "\n\t" << (*it);
    }
  output->output_f (ss.str ());
  output->output_f ("\n\n");
}

void
FTcpSender::print (double now)
{
  std::stringstream ss;
  ss << "TCP" << src_node_id << "," << src_port_id << "n" << dst_node_id << "," << dst_port_id << " t " << now << " rtt " << sink_link->get_queue_delay (now, this) + init_rtt	// rtt.get_state(now)
    << " r " << thr.get_state (now) << std::endl;
  output->output_f (ss.str ());
}

void
FTcpSender::print_avg ()
{
  double avg_thr_ = avg_thr.get_avg ();
  double avg_loss_ = avg_loss.get_avg ();
  double avg_rtt_ = avg_rtt.get_avg ();

  if (output)
    {
      std::stringstream ss;
      ss << "TCP" << this->get_flow_id ()
	<< " n " << flow_num
	<< " r " << avg_thr_
	<< " loss " << avg_loss_
	<< " rtt " << avg_rtt_ << " w " << avg_thr_ * avg_rtt_ << std::endl;
      output->output_f (ss.str ());
    }
}

double
FTcpSender::thr_derivs (double time, double thr_now)
{
  double rtt_now = sink_link->get_queue_delay (time, this) + init_rtt;
  double thr_before_rtt = thr.get_state (time - prop_delay);
  double thr_of_ack = sink_link->get_rate (time, this) / flow_num;

#if __TIMEOUT
  double pto = std::min (1.0, 3.0 / (thr_now * rtt_now));

  return thr_of_ack / (thr_now * rtt_now * rtt_now)
    - (thr_before_rtt - thr_of_ack) * thr_now * 0.66666666666666 * (1 - pto)
    - ((1.33333333333333) * thr_now - (1.0 / rtt_now)) * (thr_before_rtt -
							  thr_of_ack) * pto;
#else
  // NOTE: if packet loss rate > 0, the first term is larger than the second term.
  return thr_of_ack / (thr_now * rtt_now * rtt_now) - (thr_before_rtt -
						       thr_of_ack) * thr_now *
    0.6666666666666666;
#endif

}
